/*
 * Access a filesystem.
 *
 * Copyright (C) 2018, 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <fsclient/client.h>
#include <systypes/fcntl.h>



static char *dirname(char *filename)
{
  char *s = strdup(filename);
  char *t = strrchr(s, '/');
  if (t != NULL)
    *t = '\0';
  return s;
}

static void read_data(file_descriptor_t *desc)
{
  char buffer[80];
  ssize_t available, limit = 80;

  available = client_read(desc, buffer, limit);

  if (!available)
    return;

  /* Consume some of the data. */

  fputs("\n", stdout);
  fwrite((void *) buffer, sizeof(char), available, stdout);
  fflush(stdout);
}

int main(int argc, char *argv[])
{
  char buffer[] = "Some data written to the file.";
  file_descriptor_t *desc1 = NULL, *descD = NULL;
  char *dir = NULL;

  if (argc < 3)
  {
    printf("Need a mode and a filename.\n");
    return 1;
  }

  dir = dirname(argv[2]);

  /* Obtain a reference to a file. */

  if (!strcmp(argv[1], "create"))
  {
    descD = client_open(dir, O_DIRECTORY, 0);

    if (descD == NULL)
    {
      printf("No such directory %s.\n", dir);
      return 1;
    }

    printf("Waiting to create a file in %s...\n", dir);

    /* NOTE: Useful for testing race conditions. */

    //sleep(2);

    /* Wait for control over the directory. */

    client_begin_first(descD);

    printf("Open file...\n");

    desc1 = client_open(argv[2], O_RDWR | O_CREAT, 0);

    if (desc1 == NULL)
    {
      printf("Could not open file.\n");
      return 1;
    }

    /* Give the observer control over the directory and proceed. */

    client_end(descD);

    printf("Wait for another client...\n");

    client_begin_first(desc1);

    printf("File size = %d\n", desc1->obj.size);

    /* Write some data to the file. */

    printf("Writing...\n");

    client_write(desc1, buffer, strlen(buffer));
    client_flush(desc1);

    printf("File size = %d\n", desc1->obj.size);

    /* Request data from the file. */

    printf("\nContinuing to read...\n");

    read_data(desc1);

    printf("\nReading from start...\n");

    client_seek(desc1, 0, SEEK_SET);
    read_data(desc1);

    printf("\nWriting to start...\n");

    client_seek(desc1, 0, SEEK_SET);
    client_write(desc1, "More", strlen("More"));

    printf("File size = %d\n", desc1->obj.size);

    printf("\nContinuing to read...\n");

    read_data(desc1);
    client_flush(desc1);

    printf("\nYielding to the other process...\n");

    client_yield(desc1);

    printf("\nReading from start...\n");

    client_seek(desc1, 0, SEEK_SET);
    read_data(desc1);

    printf("\nWriting before the end...\n");

    client_seek(desc1, -5, SEEK_END);
    client_write(desc1, "text!", strlen("text!"));

    printf("\nReading from start...\n");

    client_seek(desc1, 0, SEEK_SET);
    read_data(desc1);

    client_flush(desc1);
  }
  else if (!strcmp(argv[1], "observe"))
  {
    descD = client_open(dir, O_DIRECTORY, 0);

    if (descD == NULL)
    {
      printf("No such directory %s.\n", dir);
      return 1;
    }

    printf("Waiting for the file to be created in %s...\n", dir);

    /* Wait for control over the directory. */

    client_begin(descD);

    printf("Try to open the file...\n");

    desc1 = client_open(argv[2], O_RDONLY, 0);

    /* Give the file creator control over the directory. */

    if (desc1 == NULL)
    {
      printf("Could not open file.\n");
      return 1;
    }

    client_end(descD);

    printf("Wait for the other client...\n");

    client_begin(desc1);

    printf("File size = %d\n", desc1->obj.size);

    printf("\nReading from start...\n");

    client_seek(desc1, 0, SEEK_SET);
    read_data(desc1);

    printf("\nYielding to the other process...\n");

    client_yield(desc1);

    printf("File size = %d\n", desc1->obj.size);

    printf("\nReading from start...\n");

    client_seek(desc1, 0, SEEK_SET);
    read_data(desc1);
  }

  /* Close files explicitly, but not doing so will cause their closure, anyway. */

  printf("\nClosing...\n");

  if (desc1 != NULL)
    client_close(desc1);

  printf("\nClosed.\n");

  free(dir);
  return 0;
}
